"""
Unit tests for megago.

Usage: python -m unittest -v megago_test
"""

import matplotlib
import unittest
from io import StringIO
# pylint: disable=no-name-in-module
from megago.megago import read_input, is_go_term, plot_similarity


class TestIsStringContaingGo(unittest.TestCase):
    """Unit tests for is_str_containing_go_term"""

    def do_test(self, input_str, expected):
        result = is_go_term(input_str)
        self.assertEqual(expected, result)

    def test_plain_go_string(self):
        expected = True
        self.do_test("GO:0005488", expected)

    def test_non_go_string(self):
        expected = False
        self.do_test("wasd", expected)

    def test_lower_case_go_string(self):
        expected = True
        self.do_test("go:0005488", expected)

    def test_short_too_short_go_string(self):
        expected = False
        self.do_test("go:000588", expected)

    def test_short_too_long_go_string(self):
        expected = False
        self.do_test("go:00058890", expected)

    def test_multiple_go_terms(self):
        expected = False
        self.do_test("GO:1234567;GO:1234568", expected)


class TestReadInput(unittest.TestCase):
    '''Unit tests for read_input'''

    def do_test(self, input_str, expected):
        "Wrapper function for testing read_input"
        result = read_input(StringIO(input_str))
        self.assertEqual(expected, result)

    def test_zero_byte_input(self):
        "Test input containing zero bytes"
        expected = [], [], []
        self.do_test('', expected)

    def test_valid_go_terms(self):
        string = """id1,GO:0005488,GO:0048037;GO:0036094
id2,GO:0098581,GO:0032490;GO:0098543
id3,GO:0050789,GO:0030155;GO:0031341"""
        expected = (["id1", "id2", "id3"],
                    [['GO:0005488'], ['GO:0098581'], ['GO:0050789']],
                    [['GO:0048037', 'GO:0036094'],
                     ['GO:0032490', 'GO:0098543'],
                     ['GO:0030155', 'GO:0031341']])
        self.do_test(string, expected)

    def test_second_tool_no_go_terms(self):
        string = """id1,GO:0098581,
id2,GO:0050789,GO:0030155;GO:0031341"""
        expected = (["id1", "id2"],
                    [['GO:0098581'], ['GO:0050789']],
                    [[], ['GO:0030155', 'GO:0031341']])
        self.do_test(string, expected)

    def test_first_tool_no_go(self):
        string = """ID,set1,set2
id1,,GO:0070279"""
        expected = (["id1"],
                    [[]],
                    [['GO:0070279']])
        self.do_test(string, expected)

    def test_no_header_first_tool_no_go(self):
        string = """id1,,GO:0070279"""
        expected = (["id1"],
                    [[]],
                    [['GO:0070279']])
        self.do_test(string, expected)


class TestPlotSimilarity(unittest.TestCase):
    '''Unit tests for plot_similarity'''

    def do_test(self, list_similarity_values, expected_exception=None):
        "Wrapper function for testing plot_similarity"
        if expected_exception:
            with self.assertRaises(expected_exception):
                plot_similarity(list_similarity_values)
        else:
            result = plot_similarity(list_similarity_values)
            self.assertTrue(isinstance(result, matplotlib.figure.Figure))

    def test_single_val(self):
        lst = [1]
        self.do_test(lst)

    def test_multi_val(self):
        lst = [3,5,9.]
        self.do_test(lst)

    def test_mixed_types(self):
        lst = [3,"wasd",5]
        self.do_test(lst, ValueError)


if __name__ == '__main__':
    unittest.main()



import json
from goatools.anno.idtogos_reader import IdToGosReader
from goatools.obo_parser import GODag
from goatools.semantic import TermCounts
import os
from . import DATA_DIR, JSON_INDEXED_FILE_PATH, UNIPROT_ASSOCIATIONS_FILE_PATH


def intialize_term_counts():
    go_freq_dict = dict()
    go_dag = GODag(os.path.join(DATA_DIR, "go-basic.obo"))

    associations = IdToGosReader(UNIPROT_ASSOCIATIONS_FILE_PATH, godag=go_dag).get_id2gos('all')
    term_counts = TermCounts(go_dag, associations)
    for i in go_dag.values():
        go_freq_dict[i.id] = term_counts.get_count(i.id)
    # write frequency dict to JSON file
    with open(JSON_INDEXED_FILE_PATH, 'w') as json_file:
        json.dump(go_freq_dict, json_file)



"""Mega-GO - Calculate semantic distance for sets of Gene Ontology (GO) terms.

Read a file that contains GO terms and print their semantic similarity.
"""

import argparse
import numbers
import seaborn as sns
import sys
import logging
import pkg_resources
import re
import os

from goatools.obo_parser import GODag
from progress.bar import IncrementalBar

from .constants import GO_DAG_FILE_PATH
from .precompute_highest_ic import get_highest_ic
from .metrics import compute_bma_metric
from .precompute_frequency_counts import get_frequency_counts



EXIT_COMMAND_LINE_ERROR = 2
EXIT_FASTA_FILE_ERROR = 3
DEFAULT_MIN_LEN = 0
DEFAULT_VERBOSE = False
HEADER = 'DOMAIN,SIMILARITY'
PROGRAM_NAME = "megago"
GO_DOMAINS = [
    "biological_process",
    "cellular_component",
    "molecular_function"
]
# How many items should be present in a sample before we compute metrics in parallel?
PARALLEL_TRESHOLD = 200

try:
    PROGRAM_VERSION = pkg_resources.require(PROGRAM_NAME)[0].version
except pkg_resources.DistributionNotFound:
    PROGRAM_VERSION = "undefined_version"


def is_go_term(candidate):
    """ Returns true if a given string could be a valid GO-term identifier. This function only checks if the string
    satisfies the formatting of a GO-term identifier, not if it actually exists in the ontology!

    Parameters
    ----------
    candidate : str
        A string for which we should verify if it's a valid GO-identifier.

    Returns
    -------
    bool
    """
    regex = re.compile(r"^go:\d{7}$", re.IGNORECASE)
    return bool(regex.match(candidate))


def read_input(in_file, header=True):
    """Read and return all GO terms that are found in an open file.

    Parameters
    ----------
    in_file : an open file object
    header : bool, optional
        If the file contains a header that should be stripped

    Returns
    -------
    A list with all GO-term id's that are present in the given file.

    """
    if header:
        next(in_file)
    return [line.rstrip() for line in in_file]


def parse_args():
    """ Parse command line arguments. This function will exit the program on a command line error!

    Returns
    -------
    Options object with command line argument values as attributes.
    """
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument('--version',
                        action='version',
                        version='%(prog)s ' + PROGRAM_VERSION)
    parser.add_argument('--log',
                        metavar='LOG_FILE',
                        type=str,
                        help='record program progress in LOG_FILE')
    parser.add_argument('--verbose', '-v', action='count', default=0)
    parser.add_argument('--plot', dest="plot_file",
                        help="Draw swarm plot of calculated similarities. Filetype is automatically determined based on"
                             " extension (e.g. .png, .svg)",
                        default=None
                        )
    parser.add_argument('sample_1',
                        metavar='SAMPLE_1',
                        nargs='?',
                        type=str,
                        help='Input file for sample 1')
    parser.add_argument('sample_2',
                        metavar='SAMPLE_2',
                        nargs='?',
                        type=str,
                        help='Input file for sample 2')

    return parser.parse_args()


class LogFile(object):
    """File-like object to log text using the `logging` module."""

    def __init__(self):
        self.logger = logging.getLogger()

    def write(self, msg, level=logging.DEBUG):
        msg = msg.strip("\n")
        self.logger.log(level, msg)

    def flush(self):
        for handler in self.logger.handlers:
            handler.flush()


class RedirectStdStreams(object):
    def __init__(self, stdout=None, stderr=None):
        self._stdout = stdout or sys.stdout
        self._stderr = stderr or sys.stderr

    def __enter__(self):
        self.old_stdout, self.old_stderr = sys.stdout, sys.stderr
        self.old_stdout.flush(); self.old_stderr.flush()
        sys.stdout, sys.stderr = self._stdout, self._stderr

    def __exit__(self, exc_type, exc_value, traceback):
        self._stdout.flush(); self._stderr.flush()
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr


def split_per_domain(go_terms, go_dag):
    """ Split a list of go_terms into three different lists that correspond to the GO-domains.

    Parameters
    ----------
    go_terms : a list of strings
        a list of GO terms that need to be divided over the different GO-domains.
    go_dag : a graph that represents the Gene Ontology

    Returns
    -------
    biological_process, cellular_component, molecular_function
        Three different lists with respectively the GO-terms that belong to the biological process, cellular component
        and molecular function domains.
    """
    output = {domain: [] for domain in GO_DOMAINS}

    for go_term in go_terms:
        if go_term in go_dag:
            ns = go_dag[go_term].namespace
            output[ns].append(go_term)
        else:
            logging.warning(f"{go_term} was not found in the Gene Ontology parsed by this script.")

    return [output[domain] for domain in GO_DOMAINS]


def get_default_go_dag():
    return GODag(GO_DAG_FILE_PATH, prt=open(os.devnull, 'w'))


def run_comparison(go_list_1, go_list_2, go_dag=None, progress=None):
    """ Compute the pairwise similarity values for all rows from the given file.

    Parameters
    ----------
    go_list_1 : a list with GO-identifiers as strings
        All GO-terms present in the first sample.
    go_list_2 : a list with GO-identifiers as strings
        All GO-terms present in the second sample.
    go_dag : GODag object
        GODag object from the goatools package
    progress : function (number) => void
        is called with the current progress value (a floating point value between 0 and 1)

    Returns
    -------
    tuple
        A tuple with 3 values. These correspond to the similarity scores of biological process, cellular component and
        molecular function respectively.
    """

    freq_dict = get_frequency_counts()
    highest_ic_anc = get_highest_ic()
    if go_dag is None:
        go_dag = GODag(GO_DAG_FILE_PATH, prt=open(os.devnull, 'w'))

    split_per_domain_1 = split_per_domain(go_list_1, go_dag)
    split_per_domain_2 = split_per_domain(go_list_2, go_dag)

    output = list()

    total_comparisons = len(set(go_list_1)) * len(set(go_list_2))
    done = 0

    def progress_reporter(batch_size):
        nonlocal done
        if progress:
            done += batch_size
            progress(done / total_comparisons)

    for i in range(len(GO_DOMAINS)):
        output.append(
            compute_bma_metric(
                split_per_domain_1[i],
                split_per_domain_2[i],
                freq_dict,
                highest_ic_anc,
                progress_reporter,
                similarity_method="lin"
            )
        )

    if progress:
        progress(1)

    return tuple(output)


def find_non_existing_terms(go_list, go_dag):
    """ Checks if each of the given terms from the go_list are present in the GO-dag. Returns a list with all terms that
    are not present.

    Parameters
    ----------
    go_list : a list with GO-identifiers as strings.
        Every identifier is looked up in the go_dag and is added to the output if it does not exist.
    go_dag: A valid GO graph.

    Returns
    -------
    set
        A set with all GO-identifiers from the go_list that are not present in the given go_dag.
    """
    return set(x for x in go_list if not is_go_term(x) or x not in go_dag)


def plot_similarity(list_similarity_values):
    l_is_number = [isinstance(x, numbers.Number) for x in list_similarity_values]
    if not all(l_is_number):
        raise ValueError(f"List contains non numeric values: {list_similarity_values}")
    ax = sns.swarmplot(x=list_similarity_values)
    fig = ax.get_figure()
    return fig


def process(options):
    # The GO-terms that need to be compared can be given as a CSV-file or inline in the command as a ";" delimited
    # string.
    if re.match(".*\.[^.]+$", options.sample_1):
        logging.info("Processing sample 1 from %s", options.sample_1)
        go_list_1 = read_input(open(options.sample_1, 'r'))
    else:
        go_list_1 = options.sample_1.split(';')

    if re.match(".*\.[^.]+$", options.sample_2):
        logging.info("Processing sample 2 from %s", options.sample_2)
        go_list_2 = read_input(open(options.sample_2, 'r'))
    else:
        go_list_2 = options.sample_2.split(';')

    results = run_comparison(go_list_1, go_list_2)

    print(HEADER)
    lines = [HEADER]
    for idx, domain in enumerate(GO_DOMAINS):
        line = f"{domain},{results[idx]}"
        print(line)
        lines.append(line)

    csv_table_string = "\n".join(lines)
    list_similarity_values = []
    for l in csv_table_string.split("\n")[1:]:
        list_similarity_values.append(float(l.split(",")[1]))
    if options.plot_file:
        figure = plot_similarity(list_similarity_values)
        figure.savefig(options.plot_file)


def init_logging(log_filename, verbose):
    """Initialise the logging facility, and write log statement
    indicating the program has started, and also write out the
    command line from sys.argv

    Parameters
    ----------
    log_filename : str
        name of the log file to write to or None, if is None, log output will go tto stderr
    verbose : int
        increase verbosity level. (default is WARNING, 1->INFO, 2->DEBUG, >=3->NOTSET)

    Returns
    -------
    None

    """

    verbosity = 30 - verbose * 10
    if verbosity < 0:
        verbosity = 0

    args = {"level": verbosity,
            "filemode": 'w',
            "format": '%(asctime)s %(levelname)s - %(message)s',
            "datefmt": "%Y-%m-%dT%H:%M:%S%z"}

    if log_filename is None:
        args["stream"] = sys.stderr
    else:
        args["filename"] = log_filename

    logging.basicConfig(**args)
    logging.info('program started')
    logging.info('command line: %s', ' '.join(sys.argv))


def main():
    options = parse_args()
    init_logging(options.log, options.verbose)
    process(options)
    logging.info("Done!")


if __name__ == "__main__":
    main()



# This script precomputes all comparison values for the complete set of GO-terms. We only calculate and store the
# comparison value of two GO-terms in the same namespace, as terms from different namespaces cannot be compared with
# each other.

import math
import json
import os
import concurrent.futures

from goatools.obo_parser import GODag
from progress.bar import IncrementalBar

from .constants import GO_DAG_FILE_PATH, HIGHEST_IC_FILE_PATH
from .precompute_frequency_counts import get_frequency_counts
from .metrics import get_ic_of_most_informative_ancestor

# How large should the chunks be in which the list of terms for which ic needs to be computed should be divided?
CHUNK_SIZE = 500

# How many processes can be used simultaneously at maximum? (Set to none for default)
PROCESSES = None


def _do_compute_highest_inc(terms):
    term_counts = get_frequency_counts()
    go_dag = GODag(GO_DAG_FILE_PATH, prt=open(os.devnull, 'w'))
    return {term: get_ic_of_most_informative_ancestor(term, term_counts, go_dag) for term in terms}


def compute_highest_inc_parallel(terms):
    """ Compare all values from the given terms set in parallel by using up to PROCESSES processes simultaneously.
    Params
    ------
    terms: A list with GO-terms for which the information content should be precomputed.
    """
    print("Start precomputations of the highest_inc_anc for all GO-terms.")

    term_len = len(terms)
    highest_ic_anc = dict()

    amount_of_chunks = math.ceil(term_len / CHUNK_SIZE)

    # Effectively compute the comparisons
    with concurrent.futures.ProcessPoolExecutor(max_workers=PROCESSES) as executor:
        bar = IncrementalBar('Processing', max=amount_of_chunks, suffix='%(percent)d%% - Elapsed: %(elapsed)ds - Remaining: %(eta)ds')
        for result in executor.map(_do_compute_highest_inc, (terms[i:i + CHUNK_SIZE] for i in range(0, term_len, CHUNK_SIZE))):
            bar.next()
            highest_ic_anc.update(result)
        bar.finish()

    with open(HIGHEST_IC_FILE_PATH, 'w') as json_file:
        json.dump(highest_ic_anc, json_file)


def get_highest_ic():
    if not os.path.isfile(HIGHEST_IC_FILE_PATH):
        go_dag = GODag(GO_DAG_FILE_PATH, prt=open(os.devnull, 'w'))
        compute_highest_inc_parallel(list(go_dag.keys()))

    ic_file = open(HIGHEST_IC_FILE_PATH, 'r')
    highest_ic_anc = json.load(ic_file)
    ic_file.close()

    return highest_ic_anc


if __name__ == "__main__":
    get_highest_ic()



import pkg_resources
import os

# Main directory that contains all resources for this module.
DATA_DIR = pkg_resources.resource_filename(__name__, 'resources')

# File that contains the precomputed frequency counts.
FREQUENCY_COUNTS_FILE_PATH = os.path.join(DATA_DIR, "frequency_counts_uniprot.json")

# Serialized version of the Direct Acyclic Graph of all GO-terms. Can be downloaded from
# http://geneontology.org/docs/download-ontology/
GO_DAG_FILE_PATH = os.path.join(DATA_DIR, "go-basic.obo")

# File that contains the UniProt associations at a specific moment in time (SwissProt)
UNIPROT_ASSOCIATIONS_FILE_PATH = os.path.join(DATA_DIR, "associations-swissprot.tab")

# File that contains the precomputed information content values
HIGHEST_IC_FILE_PATH = os.path.join(DATA_DIR, "highest_ic_uniprot.json")

NAN_VALUE = float('nan')



"""Create GO-Dag plots."""

__copyright__ = "Copyright (C) 2016-2018, DV Klopfenstein, H Tang. All rights reserved."
__author__ = "DV Klopfenstein"

from goatools.cli.gosubdag_plot import PlotCli


def run():
    """Create GO-Dag plots."""
    PlotCli().cli()


if __name__ == '__main__':
    run()

# Copyright (C) 2016-2018, DV Klopfenstein, H Tang. All rights reserved.


import json
import os

from goatools.anno.idtogos_reader import IdToGosReader
from goatools.obo_parser import GODag
from goatools.semantic import TermCounts
from progress.bar import IncrementalBar

from .constants import FREQUENCY_COUNTS_FILE_PATH, UNIPROT_ASSOCIATIONS_FILE_PATH, GO_DAG_FILE_PATH


def _precompute_term_frequencies():
    print("Start precomputations of term frequencies...")
    go_freq_dict = dict()
    go_dag = GODag(GO_DAG_FILE_PATH, prt=open(os.devnull, 'w'))

    associations = IdToGosReader(UNIPROT_ASSOCIATIONS_FILE_PATH, godag=go_dag).get_id2gos('all')
    term_counts = TermCounts(go_dag, associations)

    for i in go_dag.values():
        go_freq_dict[i.id] = term_counts.get_count(i.id)
        for alt_id in i.alt_ids:
            go_freq_dict[alt_id] = term_counts.get_count(i.id)
    # write frequency dict to JSON file
    with open(FREQUENCY_COUNTS_FILE_PATH, 'w') as json_file:
        json.dump(go_freq_dict, json_file)


def get_frequency_counts():
    """ This function precomputes the term frequency counts if these are outdated or not present. If they are present and
    valid, it will directly return the frequency counts.

    Returns
    -------
    A dictionary that maps each GO-term onto it's frequency counts.
    """
    if not os.path.isfile(FREQUENCY_COUNTS_FILE_PATH):
        _precompute_term_frequencies()

    frequency_dict = json.load(open(FREQUENCY_COUNTS_FILE_PATH))
    return frequency_dict


if __name__ == "__main__":
    get_frequency_counts()



#!/usr/bin/env python

from setuptools import setup

LONG_DESCRIPTION = \
'''
Mega-GO calculates the similarity between GO terms with the relevance semantic similarity (sim<sub>Rel</sub>) metric.
The program reads one or more input FASTA files.
For each file it computes a variety of statistics, and then
prints a summary of the statistics as output.

The goal is to provide a solid foundation for new bioinformatics command line tools,
and is an ideal starting place for new projects.'''


setup(
    name='megago',
    version='0.4.1.2020.05',
    author='Henning Schiebenhoefer',
    author_email='henning.schiebenhoefer@posteo.de',
    packages=['megago'],
    package_dir={'megago': 'megago'},
    package_data={'megago': [
        "resources/associations-uniprot-sp-20200116.tab",
        "resources/go-basic.obo",
        "resources/frequency_counts_uniprot.json",
        "resources/highest_ic_uniprot.json"
    ]},
    entry_points={
        'console_scripts': ['megago = megago.megago:main']
    },
    url='https://github.com/MEGA-GO/megago',
    license='LICENSE',
    description=('Calculate semantic distance for sets of Gene Ontology terms'),
    long_description=(LONG_DESCRIPTION),
    install_requires=[
        "goatools",
        "seaborn",
        "progress"
    ],
)



import math
import concurrent.futures
import os

from goatools.obo_parser import GODag
from goatools.semantic import deepest_common_ancestor
from goatools.gosubdag.gosubdag import GoSubDag

from .constants import NAN_VALUE, GO_DAG_FILE_PATH

PROCESSES = 6
CHUNK_SIZE = 50


def get_frequency(go_id, term_counts, go_dag):
    """get the relative frequency of go_id in it's respective namespace.

    The number of occurrences of the provided go_id and all of its children is divided by the total number of term
    occurrences in the associated GO namespace.

    Parameters
    ----------
    go_id : str
        gene ontology ID that the relative frequency should be calculated for
    term_counts : dict
        dictionary: key: GO terms, values: number of occurrences of GO term and its children in body of evidence
    go_dag : GODag object
        GODag object from the goatools package

    Returns
    -------
    float
    """
    go_term = go_dag[go_id]
    namespace = go_term.namespace
    if namespace == 'molecular_function':
        parent_count = term_counts.get('GO:0003674')
    elif namespace == 'cellular_component':
        parent_count = term_counts.get("GO:0005575")
    else:
        parent_count = term_counts.get('GO:0008150')

    return float(term_counts.get(go_id, 0)) / parent_count


def get_info_content(go_id, term_counts, go_dag):
    """get information content of go_id

    The information content is calculated as negative natural logarithm of the relative frequency of go_id.

    Parameters
    ----------
    go_id : str
        gene ontology ID that the relative frequency should be calculated for
    term_counts : dict
        dictionary: key: GO terms, values: number of occurrences of GO term and its children in body of evidence
    go_dag : GODag object
        GODag object from the goatools package

    Returns
    -------
    float
        if relative frequency of go_id == 0: 0
        else: negative natural logarithm of the relative frequency of go_id

    """
    freq = get_frequency(go_id, term_counts, go_dag)
    if freq == 0:
        return 0
    return 0.0 - math.log(freq)


def get_ic_of_most_informative_ancestor(id, term_counts, go_dag):
    """get the information content of the go_id parent with the highest information content.

    Parameters
    ----------
    go_id : str
        GO term
    term_counts : dict
        dictionary: key: GO terms, values: number of occurrences of GO term and its children in body of evidence
    go_dag : GODag object
        GODag object from the goatools package


    Returns
    -------
    float

    """
    if term_counts.get(id, 0) > 0:
        return 0
    gosubdag_r0 = GoSubDag([id], go_dag, prt=None)
    if id in gosubdag_r0.rcntobj.go2ancestors:
        P = gosubdag_r0.rcntobj.go2ancestors[id]
        max_ic = 0
        for i in P:
            ic = get_info_content(i, term_counts, go_dag)
            if max_ic < ic:
                max_ic = ic
        return max_ic
    else:
        return 0


def get_deepest_common_ancestor(id1, id2, go_dag):
    return deepest_common_ancestor([id1, id2], go_dag)


def rel_metric(c1, c2, go_dag, term_counts, highest_ic_anc):
    """calculate semantic similarity of the GO terms id1 and id2 using the rel metric

    Formula of the metric: (2 * info_content(mica) * (1 - freq(mica))) / (info_content(go_id1) + info_content(go_id2))
    where mica is the most informative common ancestor of go_id1 and go_id2.
    Metric is implemented according to: Schlicker, A., Domingues, F.S., Rahnenführer, J. et al. A new measure for
    functional similarity of gene products based on Gene Ontology. BMC Bioinformatics 7, 302 (2006)
    doi:10.1186/1471-2105-7-302


    Parameters
    ----------
    c1 : str
        GO term
    c2 : str
        GO term
    go_dag : GODag object
        GODag object from the goatools package
    term_counts : dict
        dictionary: key: GO terms, values: number of occurrences of GO term and its children in body of evidence

    Returns
    -------
    float
        if go_id1 and go_id2 are from different GO namespaces or either of them misses in the go_dag: NAN_VALUE
        else: rel metric

    """

    if (c1 not in go_dag) or (c2 not in go_dag):
        return NAN_VALUE

    go_term1 = go_dag[c1]
    go_term2 = go_dag[c2]
    if go_term1.namespace == go_term2.namespace:
        lca_goid = get_deepest_common_ancestor(c1, c2, go_dag)
        freq = get_frequency(lca_goid, term_counts, go_dag)
        info_content_lca = get_info_content(lca_goid, term_counts, go_dag)
        info_content1 = get_info_content(c1, term_counts, go_dag)
        info_content2 = get_info_content(c2, term_counts, go_dag)
        if info_content1 == 0:
            info_content1 = highest_ic_anc[c1]
        if info_content2 == 0:
            info_content2 = highest_ic_anc[c2]
        if info_content1 + info_content2 == 0:
            return 0
        return (2 * info_content_lca * (1 - freq)) / (info_content1 + info_content2)
    else:    # if goterms are from different GO namespaces (molecular function, cellular component, biological process)
        return NAN_VALUE


def lin_metric(c1, c2, go_dag, term_counts, highest_ic_anc):
    """calculate semantic similarity of the GO terms id1 and id2 using the rel metric

    Formula of the metric: (2 * info_content(mica)) / (info_content(go_id1) + info_content(go_id2))
    where mica is the most informative common ancestor of go_id1 and go_id2.

    Metric is implemented according to: Lin, Dekang. 1998. “An Information-Theoretic Definition of Similarity.” In
    Proceedings of the 15th International Conference on Machine Learning, 296—304.


    Parameters
    ----------
    c1 : str
        GO term
    c2 : str
        GO term
    go_dag : GODag object
        GODag object from the goatools package
    term_counts : dict
        dictionary: key: GO terms, values: number of occurrences of GO term and its children in body of evidence

    Returns
    -------
    float
        if go_id1 and go_id2 are from different GO namespaces or either of them misses in the go_dag: NAN_VALUE
        else: rel metric

    """
    if (c1 not in go_dag) or (c2 not in go_dag):
        return NAN_VALUE

    go_term1 = go_dag[c1]
    go_term2 = go_dag[c2]
    if go_term1.namespace == go_term2.namespace:
        lca_goid = get_deepest_common_ancestor(c1, c2, go_dag)
        info_content_lca = get_info_content(lca_goid, term_counts, go_dag)
        info_content1 = get_info_content(c1, term_counts, go_dag)
        info_content2 = get_info_content(c2, term_counts, go_dag)
        if info_content1 == 0:
            info_content1 = highest_ic_anc[c1]
        if info_content2 == 0:
            info_content2 = highest_ic_anc[c2]
        if info_content1 + info_content2 == 0:
            return 0
        return (2 * info_content_lca) / (info_content1 + info_content2)
    else:    # if goterms are from different GO namespaces (molecular function, cellular component, biological process)
        return NAN_VALUE


def compute_similarity_method(params):
    (go_list1, go_list2, term_counts, go_dag_path, highest_ic_anc, similarity_method) = params
    go_dag = GODag(GO_DAG_FILE_PATH, prt=open(os.devnull, 'w'))
    result = dict()

    for id1 in go_list1:
        for id2 in go_list2:
            key = (id1, id2) if id1 < id2 else (id2, id1)
            result[key] = similarity_method(id1, id2, go_dag, term_counts, highest_ic_anc)
    return result


def compute_bma_metric(go_list1, go_list2, term_counts, highest_ic_anc, progress_listener=None, similarity_method="rel"):
    """calculate the best match average similarity of the two provided sets of go terms

    For each GO term in go_list1, the similarity value of the most similar term from go_list2 is picked. The sum of
    these highest similarity values is divided by the total number of GO terms in go_list1 and go_list2. The metric
    is implemented according to: Schlicker, A., Domingues, F.S., Rahnenführer, J. et al. A new measure for functional
    similarity of gene products based on Gene Ontology. BMC Bioinformatics 7, 302 (2006) doi:10.1186/1471-2105-7-302

    Parameters
    ----------
    go_list1 : iterable
        iterable, containing go term strings
    go_list2 : iterable
        iterable, containing go term strings
    term_counts : dict
        dictionary: key: GO terms, values: number of occurrences of GO term and its children in body of evidence
    highest_ic_anc : dict
        dictionary: key: GO terms, values: information content of the ancestor with the highest information content
    progress_listener: function (number) => void
        is called with comparisons that currently have been performed
    similarity_method : string
        choose between lin and rel metric. 'lin' -> lin_metric, 'rel' -> rel_metric

    Returns
    -------
    float

    """

    unique_list1 = list(set(go_list1))
    unique_list2 = list(set(go_list2))

    # Store the results of the similarity method in this dictionary
    sim_method_dict = dict()

    amount_of_chunks = math.ceil(len(unique_list1) / CHUNK_SIZE)

    sim_func = None
    if similarity_method == "lin":
        sim_func = lin_metric
    elif similarity_method == "rel":
        sim_func = rel_metric
    else:
        raise AttributeError(f"similarity_method must be in ['lin', 'rel'] but is {similarity_method}")

    with concurrent.futures.ProcessPoolExecutor(max_workers=PROCESSES) as executor:
        chunks = []
        for process in range(amount_of_chunks):
            if process == amount_of_chunks - 1:
                chunks.append((unique_list1[CHUNK_SIZE * process:], unique_list2, term_counts, GO_DAG_FILE_PATH, highest_ic_anc, sim_func))
            else:
                chunks.append((unique_list1[CHUNK_SIZE * process: CHUNK_SIZE * (process + 1)], unique_list2, term_counts, GO_DAG_FILE_PATH, highest_ic_anc, sim_func))
        for result in executor.map(compute_similarity_method, chunks):
            progress_listener(len(result))
            sim_method_dict.update(result)


    summation_set12 = 0.0
    summation_set21 = 0.0

    for id1 in go_list1:
        max_value = 0.0
        for id2 in go_list2:
            key = (id1, id2) if id1 < id2 else (id2, id1)
            value = sim_method_dict[key]
            if value > max_value:
                max_value = value
        if len(go_list2) == 0:
            max_value = NAN_VALUE
        summation_set12 += max_value
    for id2 in go_list2:
        max_value = 0.0
        for id1 in go_list1:
            key = (id1, id2) if id1 < id2 else (id2, id1)
            value = sim_method_dict[key]
            if value > max_value:
                max_value = value
        if len(go_list1) == 0:
            max_value = NAN_VALUE
        summation_set21 += max_value

    if (len(go_list1) + len(go_list2)) == 0:
        bma = 0
    else:
        bma = (summation_set12 + summation_set21) / (len(go_list1) + len(go_list2))
    return bma
